import json
import os
import re
import sys
import time
from datetime import datetime
import pandas as pd
import requests
from tqdm import tqdm

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..data.Stock import StockData, ConstituentStocks
    from ..log.log import hide, show, slog, sprint
except:
    from log.log import hide, show, slog, sprint
    from data.Stock import StockData, ConstituentStocks


class IndustryAnalysis(object):
    '''
    同业比较
    '''

    def __init__(self, names=['贵州茅台', '隆基股份']):
        self.names = names
        self.codes = {i: j for i, j in stock_pair(
            names=names).items()}.values()
        self.url = 'http://f10.eastmoney.com/IndustryAnalysis/IndustryAnalysisAjax?'
        self.headers = {
            "Host": "f10.eastmoney.com",
            "Referer": "http://f10.eastmoney.com/IndustryAnalysis/Index?type=web&code=SZ300059",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.128 Safari/537.36 Edg/89.0.774.77",
            "X-Requested-With": "XMLHttpRequest"
        }

    def params_list(self, codes):
        params_list = []
        for code in codes:
            params_list.append({"code": code[0:2].upper()+code[3:],
                                "icode": "447"})
        return params_list

    def request(self, params):
        response = requests.get(
            self.url, headers=self.headers, params=params, timeout=100)
        response.encoding = response.apparent_encoding
        return json.loads(response.text)

    def get_data(self):
        data_list = []
        for params in tqdm(self.params_list(codes=self.codes)):
            data_list.append(self.request(params=params))
        return data_list

    @property
    def info(self):
        '''
        凡人，休得僭越！
        '''
        sprint('Getting industry analysis data...')
        industry_info = []
        growth_info = []
        valuation_info = []
        dupont_info = []
        market_size = []
        i = 0
        for data in self.get_data():
            industry_info.append({self.names[i]: data['hyzx']})  # 行业资讯
            growth_info.append(data['czxbj']['data'])  # 成长性比较
            valuation_info.append({self.names[i]: data['gzbj']['data']})  # 估值
            dupont_info.append({self.names[i]: data['dbfxbj']['data']})  # 杜邦
            market_size.append(
                {self.names[i]+'——'+'按总市值排名': data['gsgmzsz']})  # 总市值
            market_size.append(
                {self.names[i]+'——'+'按流通市值排名': data['gsgmltsz']})  # 流通市值
            market_size.append(
                {self.names[i]+'——'+'按营业收入排名': data['gsgmyysr']})  # 营业收入
            market_size.append(
                {self.names[i]+'——'+'按净利润排名': data['gsgmjlr']})  # 净利润
            i += 1
        return {
            'industry_info': industry_info,
            'growth_info': growth_info,
            'valuation_info': valuation_info,
            'dupont_info': dupont_info,
            'market_size': market_size,
        }

    def industry_info(self):
        '''
        企业资讯
        '''
        lists = []
        for i in self.info['industry_info']:
            for j, k in i.items():
                for l in k:
                    lists.append([j, l['date'], l['title']])
        return pd.DataFrame(lists, columns=['stock', 'date', 'advisory'])

    def growth_info(self):
        '''
        成长性比较
        '''
        lists = []
        for i in self.info['growth_info']:
            lists.append([i[0]['jc']])
            for j in i:
                lists.append(j.values())
        T1 = ['基本每股收益增长率(%)', '营业收入增长率(%)', '净利润增长率(%)']
        T2 = ['3年复合', '19A', 'TTM', '20E', '21E', '22E']
        columns = ['排名', '代码', '简称']
        for t1 in T1:
            for t2 in T2:
                columns.append(t2+'--'+t1)
        return pd.DataFrame(lists, columns=columns)

    def valuation_info(self):
        '''
        估值比较\n
        (1)MRQ市净率=上一交易日收盘价/最新每股净资产\n
        (2)市现率①=总市值/现金及现金等价物净增加额\n
        (3)市现率②=总市值/经营活动产生的现金流量净额
        '''
        columns = ['排名', '代码', '简称', 'PEG']
        T = {'市盈率': ['19A', 'TTM', '20E', '21E', '22E'],
             '市销率': ['19A', 'TTM', '20E', '21E', '22E'],
             '市净率': ['19A', 'MRQ'],
             '市现率①': ['19A', 'TTM'],
             '市现率②': ['19A', 'TTM'],
             'EV/EBITDA': ['19A', 'TTM']}
        for i, j in T.items():
            for k in j:
                columns.append(k+'--'+i)
        lists = []
        for i in self.info['valuation_info']:
            lists.append(i.keys())
            for j in i.values():
                for k in j:
                    lists.append(k.values())
        return pd.DataFrame(lists, columns=columns)

    def dupont_info(self):
        '''
        杜邦分析比较
        '''
        T1 = ['ROE(%)', '净利率(%)', '总资产周转率(%)', '权益乘数(%)']
        T2 = ['3年平均', '17A', '18A', '19A']
        columns = ['排名', '代码', '简称']
        for t1 in T1:
            for t2 in T2:
                columns.append(t2+'--'+t1)
        lists = []
        for i in self.info['dupont_info']:
            lists.append(i.keys())
            for j in i.values():
                for k in j:
                    lists.append(k.values())
        return pd.DataFrame(lists, columns=columns)

    def market_size(self):
        '''
        公司规模
        '''
        columns = ['排名', '代码', '简称',
                   '总市值(元)', '流通市值(元)', '营业收入(元)', '净利润(元)', '报告期']
        lists = []
        for i in self.info['market_size']:
            lists.append(i.keys())
            for j in i.values():
                for k in j:
                    lists.append(k.values())
        return pd.DataFrame(lists, columns=columns)


class FinancialStatements(object):
    '''
    会计报表和主要财务指标
    '''

    def __init__(self, names=['贵州茅台', '隆基股份']):
        self.names = names
        self.code_dict = {i: j[3:] for i, j in stock_pair(names=names).items()}
        self.main_urls = [
            f'http://quotes.money.163.com/service/zycwzb_{i}.html?type=report' for i in self.code_dict.values()]
        self.statement_urls = [
            f'http://quotes.money.163.com/service/cwbbzy_{i}.html' for i in self.code_dict.values()]

    def request(self, urls):
        '''
        大胆！不准看！
        '''
        sprint('Getting data...')
        result_list = []
        n = 0
        for url in tqdm(urls):
            response = requests.get(url, timeout=100)
            response.encoding = response.apparent_encoding
            data = response.text.replace(
                '\r', '').replace('\t', '').split('\n')
            df = pd.DataFrame(
                [i.split(',') for i in data if i != '']).set_index(0).T
            df['名称'] = self.names[n]
            n += 1
            result_list.append(df)
        df = pd.concat(result_list)
        return df

    def summary(self):
        '''财务报表摘要'''
        return self.request(self.statement_urls)

    def balance(self):
        '''资产负债表'''
        return self.request([i.replace('cwbbzy', 'zcfzb') for i in self.statement_urls])

    def income(self):
        '''利润表'''
        return self.request([i.replace('cwbbzy', 'lrb') for i in self.statement_urls])

    def cash_flow(self):
        '''现金流量表'''
        return self.request([i.replace('cwbbzy', 'xjllb') for i in self.statement_urls])

    def main_info(self):
        '''主要财务指标'''
        return self.request(self.main_urls)

    def profit_info(self):
        '''盈利能力指标'''
        return self.request([i+'&part=ylnl' for i in self.main_urls])

    def debt_info(self):
        '''偿债能力指标'''
        return self.request([i+'&part=chnl' for i in self.main_urls])

    def growth_info(self):
        '''成长能力指标'''
        return self.request([i+'&part=cznl' for i in self.main_urls])

    def operation_info(self):
        '''营运能力指标'''
        return self.request([i+'&part=yynl' for i in self.main_urls])

    def save(self, data, file_name='data', path='./'):
        '''
        按照名称分类保存数据在excel不同的sheet中\n
        data --> 传入本类中任意方法的返回数据\n
        file_name --> 保存的excel名称\n
        path --> 保存路径，默认在当前文件夹
        '''
        writer = pd.ExcelWriter(path+f'{file_name}.xlsx')
        for name in self.names:
            df = data[data['名称'] == name]
            del df['名称']
            df.to_excel(writer, name, index=False)
        writer.save()


if __name__ == '__main__':
    from pprint import pprint
    ia = IndustryAnalysis(names=['贵州茅台', '隆基股份'])
    print(ia.dupont_info())



import concurrent.futures
import datetime
import json
import os
import re
import sys
import time
from pprint import pprint

import baostock as bs
import pandas as pd
import pretty_errors
import requests
from tqdm import tqdm, trange

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..log.log import hide, makedir, show, slog, sprint
except:
    from log.log import hide, makedir, show, slog, sprint


def get_data(rs):
    data_list = []
    while (rs.error_code == '0') & rs.next():
        data_list.append(rs.get_row_data())
    return pd.DataFrame(data_list, columns=rs.fields)


def login():
    hide()
    bs.login()
    show()


class StockData(object):
    '''
    股票数据
    names=['贵州茅台', '隆基股份', '五粮液']
    start_date='2019-01-01'
    end_date='2020-01-01'
    adjustflag='3' --> 默认不复权
    path --> 默认缓存路径为：".\\Suluoya cache\\"，可传入False不缓存
    '''

    def __init__(self, names=['贵州茅台', '隆基股份'],
                 start_date='2019-12-01', end_date='2020-12-31',
                 frequency='d',
                 adjustflag='3',
                 path='.\\Suluoya cache\\'):
        self.names = names
        self.start_date = start_date
        self.end_date = end_date
        self.frequency = frequency
        self.adjustflag = adjustflag  # 默认不复权
        self.path = path
        if self.path:
            makedir(self.path, '')
        login()

    def logout(self):
        '''
        退出
        '''
        hide()
        bs.logout()
        show()

    def stocks_info(self):
        '''
        Return a dict containing stock names, codes and ipoDate
        {
            '贵州茅台': {'code': 'sh.600519', 'ipoDate': '2001-08-27'},
            '隆基股份': {'code': 'sh.601012', 'ipoDate': '2012-04-11'},
            ...
        }
        '''
        info = {}
        sprint('Loading stocks information...')
        for name in tqdm(self.names):
            rs = bs.query_stock_basic(code_name=name)
            stock_info = get_data(rs)
            info[name] = {'code': stock_info['code'][0],
                          'ipoDate': stock_info['ipoDate'][0]}
        if self.path:
            makedir(self.path, 'stock data')
            df_info = pd.DataFrame(info).T
            df_info['name'] = df_info.index
            df_info.to_csv(
                f'{self.path}\\stock data\\stocks_info.csv', index=False)
        return info

    def stocks_data(self):
        '''
        Return a DataFrame containing all the stocks data
        date,code,open,high,low,close,preclose,volume,amount,adjustflag,turn,tradestatus,pctChg,peTTM,psTTM,pcfNcfTTM,pbMRQ,isST
        戳下面这个链接
        http://baostock.com/baostock/index.php/A股K线数据
        '''
        if not self.path:
            stocks_info = self.stocks_info()
        else:
            try:
                stocks_info = pd.read_csv(
                    f'{self.path}\\stock data\\stocks_info.csv').set_index('name').T.to_dict()
            except:
                stocks_info = self.stocks_info()
        df_list = []
        sprint('Loading stocks data...')
        for name in tqdm(self.names):
            code = stocks_info[name]['code']
            if stocks_info[name]['ipoDate'] > self.start_date:
                sprint(
                    f"{name}'s ipo date is {stocks_info[name]['ipoDate']}, which is after {self.start_date}.")
            if self.frequency == 'd':
                rs = bs.query_history_k_data_plus(code,
                                                  'date,code,open,high,low,close,preclose,volume,amount,adjustflag,turn,tradestatus,pctChg,peTTM,psTTM,pcfNcfTTM,pbMRQ,isST',
                                                  start_date=self.start_date, end_date=self.end_date,
                                                  frequency='d', adjustflag=self.adjustflag)
            elif self.frequency == 'w':
                rs = bs.query_history_k_data_plus(code,
                                                  'date,code,open,high,low,close,volume,amount,adjustflag,turn,pctChg',
                                                  start_date=self.start_date, end_date=self.end_date,
                                                  frequency='w', adjustflag=self.adjustflag)
            elif self.frequency == 'm':
                rs = bs.query_history_k_data_plus(code,
                                                  'date,code,open,high,low,close,volume,amount,adjustflag,turn,pctChg',
                                                  start_date=self.start_date, end_date=self.end_date,
                                                  frequency='m', adjustflag=self.adjustflag)
            df = get_data(rs)
            df['name'] = name
            df_list.append(df)
        df = pd.concat(df_list).apply(pd.to_numeric, errors='ignore')
        df['date'] = pd.to_datetime(df['date'], format='%Y-%m-%d')
        if self.path:
            df.to_csv(f'{self.path}\\stock data\\stocks_data.csv', index=False)
        return df


class ConstituentStocks(StockData):
    def __init__(self):
        hide()
        bs.login()
        show()

    def stock_industry(self):
        '''
        Return a DataFrame containing all stock industry data
        戳下面这个链接
        http://baostock.com/baostock/index.php/行业分类
        '''
        return get_data(bs.query_stock_industry())

    def sz50(self):
        '''上证50'''
        return get_data(bs.query_sz50_stocks())

    def hs300(self):
        '''沪深300'''
        return get_data(bs.query_hs300_stocks())

    def zz500(self):
        '''中证500'''
        return get_data(bs.query_zz500_stocks())


def GetGoodStock(page=5):
    sprint('Getting data from http://fund.eastmoney.com/data/rankhandler.aspx ...')
    url = "http://fund.eastmoney.com/data/rankhandler.aspx"
    headers = {
        "Host": "fund.eastmoney.com",
        "Referer": "http://fund.eastmoney.com/data/fundranking.html",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36 Edg/88.0.705.63"
    }
    urls = []

    def get_urls(page):
        params = {
            "op": "ph",
            "sc": "6yzf",
            "sd": f'{time.strftime("%Y-%m-%d", time.localtime())}',
            "ed": f'{time.strftime("%Y-%m-%d", time.localtime())}',
            "pi": str(page),
            "dx": "1",
        }
        response = requests.get(url, headers=headers, params=params)
        response.encoding = response.apparent_encoding
        data = re.findall('var rankData = {datas:(.*),allRe', response.text)[0]
        data = eval(data)
        list = ['http://fund.eastmoney.com/' +
                re.findall(r'(\d*),', i)[0]+'.html' for i in data]
        for i in list:
            urls.append(i)
    for i in range(1, page+1):
        get_urls(i)

    def get_stock(url):
        df = pd.read_html(url)
        return df[5][['股票名称', '持仓占比']]

    stocks = []

    def main(url):
        stocks.append(get_stock(url))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        for url in urls:
            executor.submit(main, url)

    stock = pd.concat(stocks)
    stock['持仓占比'] = stock['持仓占比'].map(lambda x: x.replace('%', ''))
    stock = stock.replace('暂无数据', 0)
    stock['持仓占比'] = stock['持仓占比'].astype('float')
    group = stock.groupby('股票名称')
    df1 = group.mean()
    df2 = group.count()
    df1 = df1.rename(columns={'持仓占比': '平均持仓占比'})
    df2 = df2.rename(columns={'持仓占比': '出现次数'})
    df = pd.merge(df1, df2, how='outer', on='股票名称')
    df = df.sort_values(by='出现次数', ascending=False)
    return df


if __name__ == '__main__':
    cs = StockData(names=['沪深300指数', '隆基股份'], start_date='2021-01-28',
                   end_date='2021-09-28', frequency='w',
                   path=r'.\\StockData cache\\')
    test = cs.stocks_info()
    print(test)



__name__ = "Suluoya"
__author__ = 'Suluoya'
__all__ = ['access', 'pre', 'ai']

from .assess import topsis, grey_relation, entropy_weight, entropy_weight_topsis, ladder_distribution, fuzzy_synthesis
from .pre import DataPreprocessor
from .ai import exploratory_experiment, quantile_regression, linear_regression, logistic_regression



import os
import sys
import time

import pandas as pd
import pretty_errors
import PySimpleGUI as sg

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..data.Bond import BondData
    from ..data.Company import FinancialStatements, IndustryAnalysis
    from ..data.Date import GetDate
    from ..data.Stock import ConstituentStocks, GetGoodStock, StockData
    from ..log.log import hide, show, slog, sprint
    from ..stock.Kline import SyntheticIndex
    from ..stock.Markovitz import Markovitz
except:
    from data.Bond import BondData
    from data.Company import FinancialStatements, IndustryAnalysis
    from data.Date import GetDate
    from data.Stock import ConstituentStocks, GetGoodStock, StockData
    from log.log import hide, show, slog, sprint
    from stock.Kline import SyntheticIndex
    from stock.Markovitz import Markovitz


class App(object):
    '''A simple APP'''

    def __init__(self):
        sg.theme('BlueMono')

    def window_control(self, window):
        event, values = window.read()
        window.close()
        if not event:
            sys.exit()
        return values

    def path_control(self, path):
        if not path:
            raise ValueError('Please choose a folder to save the result!')

    def save(self, path, func_name, dic):
        writer = pd.ExcelWriter(f'{path}/{func_name}.xlsx')
        for name, data in dic.items():
            data.to_excel(writer, sheet_name=name)
        writer.save()
        sprint(f'Saved in {path}.')

    def markovitz_gui(self):
        func_name = 'Markovitz Portfolio'
        layout = [
            [sg.Text('Stock List')],
            [sg.Multiline('迎驾贡酒\n明微电子\n健民集团\n西部超导\n天华超净', key='stock_list')],
            [sg.Text('Start Date'), sg.Input(
                '2021-05-05', key='start_date')],
            [sg.Text('End Date'), sg.Input(
                '2021-11-05', key='end_date')],
            [sg.Text('Frequency'), sg.Combo(
                ['d', 'w', 'm'], key='frequency')],
            [sg.Text('Risk-off Rate'), sg.Input(
                '0.023467', key='no_risk_rate')],
            [sg.Text('Funds'), sg.Input(
                '4500000', key='funds')],
            [sg.FolderBrowse('Choose a folder to save data', key='path')],
            [sg.Text('Weight Tests'), sg.Input(
                0, key='weight_tests')],
            [sg.Button('Start working!')],
        ]
        window = sg.Window(func_name, layout)
        values = self.window_control(window)
        path = values['path']
        self.path_control(path)
        frequency = values['frequency']
        if frequency not in ['d', 'w', 'm']:
            raise ValueError('Please choose frequency first!!!')
        start_date = values['start_date']
        end_date = values['end_date']
        stock_list = values['stock_list'].rstrip().split('\n')
        no_risk_rate = float(eval(values['no_risk_rate']))
        funds = float(values['funds'])
        mk = Markovitz(names=stock_list, start_date=start_date,
                       end_date=end_date, frequency=frequency, no_risk_rate=no_risk_rate, funds=funds, path=path+'\\Markovitz')
        if int(values['weight_tests']) != 0:
            return mk.weight_tests(number=int(values['weight_tests']))
        opt_dict = mk.optimization()
        df_opt = pd.DataFrame(opt_dict)
        df_opt['sharpe'] = None
        df_opt['sharpe'].iloc[0] = opt_dict['sharpe']

        result_dict = mk.buy()
        df_buy = pd.DataFrame(result_dict)
        df_buy['sharpe'] = None
        df_buy['sharpe'].iloc[0] = result_dict['sharpe']
        df_buy['sum_cost'] = None
        df_buy['sum_cost'].iloc[0] = result_dict['sum_cost']
        df_buy.columns = ['shares', 'weights', 'cost', 'sharpe', 'sum_cost']
        dic = {
            'stock data': mk.data,
            'scatter data': mk.scatter_data(number=5000),
            'boundary scatter data': mk.boundary_scatter_data(number=500),
            'optimal weight': df_opt,
            'init_port': mk.init_port(),
            'buy': df_buy,
        }
        dic.update(mk.calculate())
        name = func_name+'-'+str(stock_list)+'-'+str(start_date)+'--' + \
            str(end_date)+'-'+str(no_risk_rate)+'-'+values['funds']
        self.save(path, name, dic)
        si = SyntheticIndex(
            weights=opt_dict['weights'], path=path+'\\Markovitz')
        si.draw_chart()
        mk.heatmap()
        mk.CML()

    def markovitz_work(self):
        if self.event == 'Markovitz Portfolio':
            self.window.close()
            self.markovitz_gui()

    def stock_gui(self):
        func_name = 'Stock Data'
        layout = [
            [sg.Text('Stock List')],
            [sg.Multiline('贵州茅台\n隆基股份\n五粮液', key='stock_list')],
            [sg.Text('Start Date'), sg.Input(
                '2019-01-01', key='start_date')],
            [sg.Text('End Date'), sg.Input(
                '2020-01-01', key='end_date')],
            [sg.Text('Frequency'), sg.Combo(
                ['d', 'w', 'm'], key='frequency')],
            [sg.FolderBrowse('Choose a folder to save data', key='path')],
            [sg.Button('Start working!')]
        ]
        window = sg.Window(func_name, layout)
        values = self.window_control(window)
        path = values['path']
        self.path_control(path)
        start_date = values['start_date']
        end_date = values['end_date']
        stock_list = values['stock_list'].rstrip().split('\n')
        frequency = values['frequency']
        if frequency not in ['d', 'w', 'm']:
            raise ValueError('Please choose frequency first!!!')
        sd = StockData(names=stock_list,
                       start_date=start_date,
                       end_date=end_date,
                       frequency=frequency,
                       adjustflag='3', path=False)
        df_info = pd.DataFrame(sd.stocks_info()).T.set_index('name')
        cs = ConstituentStocks()
        dic = {
            '股票代码和上市日期': df_info,
            '股票数据': sd.stocks_data(),
            '股票行业数据': cs.stock_industry(),
            '上证50': cs.sz50(),
            '沪深300': cs.hs300(),
            '中证500': cs.zz500()
        }
        name = func_name+'-'+str(stock_list)+'-' + \
            str(start_date)+'--' + str(end_date)
        self.save(path, name, dic)

    def stock_work(self):
        if self.event == 'Stock Data':
            self.window.close()
            self.stock_gui()

    def company_gui(self):
        func_name = 'Company Data'
        layout = [
            [sg.Text('Stock List')],
            [sg.Multiline('贵州茅台\n隆基股份\n五粮液', key='stock_list')],
            [sg.FolderBrowse('Choose a folder to save data', key='path')],
            [sg.Button('Start working!')]
        ]
        window = sg.Window(func_name, layout)
        values = self.window_control(window)
        path = values['path']
        self.path_control(path)
        stock_list = values['stock_list'].rstrip().split('\n')
        fs = FinancialStatements(names=stock_list)
        ia = IndustryAnalysis(names=stock_list)
        dic = {
            '财务报表摘要': fs.summary(),
            '资产负债表': fs.balance(),
            '利润表': fs.income(),
            '现金流量表': fs.cash_flow(),
            '主要财务指标': fs.main_info(),
            '盈利能力指标': fs.profit_info(),
            '偿债能力指标': fs.debt_info(),
            '成长能力指标': fs.growth_info(),
            '营运能力指标': fs.operation_info(),
            '行业资讯': ia.industry_info(),
            '成长性比较': ia.growth_info(),
            '估值比较': ia.valuation_info(),
            '杜邦分析比较': ia.dupont_info(),
            '公司规模': ia.market_size()
        }
        name = func_name+'-'+str(stock_list)
        self.save(path, name, dic)

    def company_work(self):
        if self.event == 'Company Data':
            self.window.close()
            self.company_gui()

    def bond_gui(self):
        func_name = 'Bond Data'
        layout = [
            [sg.FolderBrowse('Choose a folder to save data', key='path')],
            [sg.Button('Start working!')]
        ]
        window = sg.Window(func_name, layout)
        values = self.window_control(window)
        path = values['path']
        self.path_control(path)
        bd = BondData()
        dic = {
            '可转债': bd.convertible_bond(),
            '企业债': bd.corporate_bond(),
            '国债': bd.national_bond()
        }
        name = func_name+'-'+str(time.strftime("%Y%m%d", time.localtime()))
        self.save(path, name, dic)

    def bond_work(self):
        if self.event == 'Bond Data':
            self.window.close()
            self.bond_gui()

    def date_gui(self):
        func_name = 'Date Data'
        layout = [
            [sg.Text('Start Date'), sg.Input(
                '20190101', key='start_date')],
            [sg.Text('End Date'), sg.Input(
                str(time.strftime("%Y%m%d", time.localtime())), key='end_date')],
            [sg.FolderBrowse('Choose a folder to save data', key='path')],
            [sg.Button('Start working!')]
        ]
        window = sg.Window(func_name, layout)
        values = self.window_control(window)
        path = values['path']
        self.path_control(path)
        start_date = values['start_date']
        end_date = values['end_date']
        gd = GetDate()
        dic = {'date': gd.Date()}
        name = func_name + '-' + str(start_date)+'--' + str(end_date)
        self.save(path, name, dic)

    def date_work(self):
        if self.event == 'Date Data':
            self.window.close()
            self.date_gui()

    def work(self):

        layout = [
            [sg.Button('Markovitz Portfolio'), ],
            [sg.Button('Stock Data'), sg.Button('Company Data')],
            [sg.Button('Date Data'), sg.Button('Bond Data')],
        ]
        self.window = sg.Window('Suluoya', layout)
        self.event, self.values = self.window.read()
        while True:
            if not self.stock_work() \
                    and not self.markovitz_work() \
                    and not self.company_work() \
                    and not self.bond_work() \
                    and not self.date_work():
                sys.exit()


def app():
    app = App()
    app.work()


if __name__ == '__main__':
    app()



__name__ = "Suluoya"
__author__ = 'Suluoya'
__all__ = ['Stock', 'Bond', 'Company']

from .Stock import StockData, ConstituentStocks, GetGoodStock
from .Bond import BondData
from .Company import IndustryAnalysis, FinancialStatements
from .Fund import FundData
from .Date import GetDate


import math
import os
import random
import sys
import time
import warnings
from itertools import combinations, product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pretty_errors
import scipy.optimize as sco
import seaborn as sns
import statsmodels.api as sm
from statsmodels import regression
from tqdm import tqdm, trange

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..data.Stock import StockData
    from ..log.log import hide, makedir, progress_bar, show, slog, sprint
except:
    from data.Stock import StockData
    from log.log import hide, makedir, progress_bar, show, slog, sprint

warnings.filterwarnings('ignore')
plt.rcParams['font.sans-serif'] = ['FangSong']
plt.rcParams['axes.unicode_minus'] = False  # 正常显示负号
plt.rcParams['font.size'] = 13


class Markovitz(object):
    '''
    组合投资权重\n
    names=['贵州茅台', '隆基股份', '五粮液']\n
    start_date='2021-05-01'\n
    end_date='2021-11-01'\n
    rfr=0.023467/365\n
    funds=10000000\n
    path --> 默认缓存路径为：".\\Suluoya cache\\"，可传入False不缓存
    '''

    def __init__(self, names=['比亚迪', '阳光电源', '璞泰来', '紫光国微', '盛新锂能'],
                 start_date='2021-05-01',
                 end_date='2021-11-01',
                 frequency='d',
                 rfr=0.023467,
                 funds=10000000,
                 path='.\\Markovitz cache\\'):
        self.names = names
        self.lens = len(names)
        self.start_date = start_date
        self.end_date = end_date
        self.frequency = frequency
        self.rfr = (rfr*100) / \
            {'d': 365, 'w': 52, 'm': 30}[frequency]
        self.funds = funds
        self.path = path
        if self.path:
            makedir(self.path, '')
        sprint('Initializing...')
        if not self.path:
            sd = StockData(names=self.names, start_date=self.start_date,
                           end_date=self.end_date, frequency=self.frequency)
            self.datas = sd.stocks_data()
        else:
            try:
                self.datas = pd.read_csv(
                    f'{self.path}\\stock data\\stocks_data.csv')
            except:
                sd = StockData(names=self.names, start_date=self.start_date,
                               end_date=self.end_date, frequency=self.frequency, path=self.path)
                self.datas = sd.stocks_data()
        self.datas.index = self.datas['name']
        self.data = self.datas.reset_index(drop=True)
        self.date = list(map(lambda x: str(x)[:10], self.data.date.unique()))
        self.first_date = self.date[0]
        self.last_date = self.date[-1]
        # 第一天开盘价
        self.first_price = self.data[self.data.date == self.data.date.unique(
        )[0]][['open', 'name']].set_index('name').to_dict()['open']
        # 最后一天收盘价
        self.last_price = self.data[self.data.date == self.data.date.unique(
        )[-1]][['close', 'name']].set_index('name').to_dict()['close']
        # 每只股票最大手数
        self.max_shares_dict = {name: math.floor(
            self.funds/(shares*100)) for name, shares in self.last_price.items()}

    def weights(self, number=5000):
        '''
        生成股票随机权重
        '''
        return np.random.dirichlet(np.ones(self.lens), size=number)

    def calculate(self):
        '''
        计算收益率均值、协方差矩阵、相关系数矩阵
        '''
        data = self.data[['date', 'name', 'pctChg']]
        # 收益率 均值
        data_mean = data.groupby('name').mean().T[self.names]
        # 收益率 协方差矩阵和相关系数矩阵
        df = pd.DataFrame()
        for name in self.names:
            df[name] = list(data[data['name'] == name]['pctChg'])
        data_cov = df.cov()
        data_corr = df.corr()
        if self.path:
            makedir(self.path, 'mean,cov,corr')
            data_mean.T.to_csv(
                f'{self.path}\\mean,cov,corr\\data_mean.csv')
            data_cov.to_csv(f'{self.path}\\mean,cov,corr\\data_cov.csv')
            data_corr.to_csv(f'{self.path}\\mean,cov,corr\\data_corr.csv')
        return {'mean': data_mean, 'cov': data_cov, 'correlation': data_corr}

    def heatmap(self, show=True):
        '''
        收益率相关系数热力图
        '''
        if self.path:
            try:
                data_corr = pd.read_csv(f'{self.path}\\mean,cov,corr\\data_corr.csv').rename(
                    {'Unnamed: 0', 'correlation'}).set_index('correlation')
            except:
                data = self.data[['name', 'pctChg']]
                df = pd.DataFrame()
                for name in self.names:
                    df[name] = list(data[data['name'] == name]['pctChg'])
                data_corr = df.corr()
                data_corr.to_csv(f'{self.path}\\mean,cov,corr\\data_corr.csv')
        else:
            data = self.data[['name', 'pctChg']]
            df = pd.DataFrame()
            for name in self.names:
                df[name] = list(data[data['name'] == name]['pctChg'])
            data_corr = df.corr()
        # 画图
        plt.subplots(figsize=(9, 9))
        sns.heatmap(data_corr, annot=True, vmax=1, square=True, cmap='Purples')
        if show:
            plt.show()
        else:
            plt.savefig(f'{self.path}\\heatmap.svg', format='svg')

    def sharpe(self, weights):
        '''
        按照names顺序传入权重（权重和为1）
        '''
        data_dict = self.calculate()
        data_mean = data_dict['mean']
        data_cov = data_dict['cov']
        weights = np.array(weights)
        rate = data_mean.dot(weights.T)['pctChg']
        risk = np.sqrt(weights.dot(data_cov).dot(weights.T))
        return (self.rfr-rate)/risk  # 相反数

    def optimization(self):
        '''
        非线性规划求解最大夏普比率
        '''
        opts = sco.minimize(fun=self.sharpe,
                            # 传入股票权重，即shapre函数的参数weights
                            x0=np.ones(self.lens)/self.lens,
                            bounds=tuple((0, 1)for x in range(self.lens)),
                            constraints={'type': 'eq',
                                         'fun': lambda x: np.sum(x) - 1}
                            )
        opt_dict = {'weights': dict(
            zip(self.names, list(opts.x))), 'sharpe': -opts.fun}
        if self.path:
            df_opt = pd.DataFrame(opt_dict)
            df_opt['sharpe'] = None
            df_opt['sharpe'].iloc[0] = opt_dict['sharpe']
            df_opt.to_csv(f'{self.path}\\max sharpe and weights.csv')
        return opt_dict

    def scatter_data(self, number=5000):
        '''
        散点数据，默认生成5000个
        '''
        data_dict = self.calculate()
        data_mean = data_dict['mean']
        data_cov = data_dict['cov']
        weights = self.weights(number=number)
        # 散点DataFrame
        df_scatter = pd.DataFrame()
        # 随机权重
        df_scatter['weights'] = pd.Series(map(lambda x: str(x), weights))
        # 风险
        df_scatter['risk'] = np.sqrt(np.diagonal(
            weights.dot(data_cov).dot(weights.T)))
        # 收益率
        df_scatter['rate'] = data_mean.dot(weights.T).T['pctChg']
        # 夏普比率
        df_scatter['sharpe'] = (
            df_scatter.rate-self.rfr)/df_scatter.risk
        df_scatter = df_scatter.sort_values(by='sharpe', ascending=False)
        if self.path:
            makedir(self.path, 'scatter data')
            df_scatter.to_csv(f'{self.path}\\scatter data\\scatter_data.csv')
        return df_scatter

    def boundary_scatter_data(self, number=500):
        '''
        边界散点数据，默认生成500个
        '''
        if self.path:
            try:
                df_scatter = pd.read_csv(
                    f'{self.path}\\scatter data\\scatter_data.csv', index=False)
            except:
                df_scatter = self.scatter_data()
        else:
            df_scatter = self.scatter_data()
        data_dict = self.calculate()
        data_mean = data_dict['mean']
        data_cov = data_dict['cov']
        scatter_list = []
        sprint('Searching for boundary scatter...')
        for i in trange(number):
            random_rate = random.uniform(
                df_scatter.rate.min(), df_scatter.rate.max())
            constraints = ({'type': 'eq', 'fun': lambda weights: weights.sum()-1},
                           {'type': 'eq', 'fun': lambda weights: data_mean.dot(weights.T)['pctChg']-random_rate})
            opts = sco.minimize(fun=lambda weights: weights.dot(data_cov).dot(weights.T),
                                x0=np.ones(self.lens)/self.lens,
                                bounds=tuple((0, 1)for x in range(self.lens)),
                                constraints=constraints
                                )
            scatter_list.append([opts.x,  np.sqrt(opts.fun), random_rate])
        df_boundary_scatter = pd.DataFrame(scatter_list, columns=[
            'weights', 'risk', 'rate'])
        df_boundary_scatter['sharpe'] = (
            df_boundary_scatter.rate-self.rfr)/df_boundary_scatter.risk
        df_boundary_scatter = df_boundary_scatter.sort_values(
            by='sharpe', ascending=False)
        if self.path:
            makedir(self.path, 'scatter data')
            df_boundary_scatter.to_csv(
                f'{self.path}\\scatter data\\boundary_scatter_data.csv')
        return df_boundary_scatter

    def cml(self, show=True):
        '''
        资本市场线 & 有效边界
        '''
        if self.path:
            try:
                df_scatter = pd.read_csv(
                    f'{self.path}\\scatter data\\scatter_data.csv')
                df_boundary_scatter = pd.read_csv(
                    f'{self.path}\\scatter data\\boundary_scatter_data.csv')
            except:
                df_scatter = self.scatter_data()
                df_boundary_scatter = self.boundary_scatter_data()
            df_scatter['boundary'] = False
            df_boundary_scatter['boundary'] = True
            pd.concat([df_scatter, df_boundary_scatter]).to_csv(
                f'{self.path}\\scatter data\\all_scatter_data.csv')
        else:
            df_scatter = self.scatter_data()
            df_boundary_scatter = self.boundary_scatter_data()

        max_sharpe = self.optimization()['sharpe']
        sprint(f'max sharpe: {max_sharpe}')
        plt.style.use('seaborn-paper')
        plt.scatter(df_scatter.risk, df_scatter.rate,
                    s=10, marker=".", c='b')
        plt.scatter(df_boundary_scatter.risk, df_boundary_scatter.rate,
                    s=10, marker=".", c='r')
        plt.axline(xy1=(0, self.rfr), slope=max_sharpe, c='m')
        plt.xlim(df_scatter.risk.min()*0.8, df_scatter.risk.max()*1.2)
        plt.ylim(df_scatter.rate.min()*0.8, df_scatter.rate.max()*1.2)
        plt.xlabel('Risk')
        plt.ylabel('Yield')
        if show:
            plt.show()
        else:
            plt.savefig(f'{self.path}\\cml.svg', format='svg')
        return pd.concat([df_scatter, df_boundary_scatter])

    def exam(self, shares_dict={'迎驾贡酒': 1, '明微电子': 2, '健民集团': 3}):
        '''
        检验：
        0<=手数<=最大手数
        总成本<=fund
        返回sharpe或0
        '''
        for name, number in shares_dict.items():
            if number > self.max_shares_dict[name]:
                return 0  # 检验不符合
        # if 0 in shares_dict.values():
        #     return 0  # 检验不符合
        weights = np.array([self.last_price[name]*100*shares_dict[name]
                           for name in self.names])
        if weights.sum() > self.funds:
            return 0  # 检验不符合
        weights = weights/weights.sum()
        return -self.sharpe(weights=weights)

    def init_port(self):
        '''
        初始临近整数组合 --> DataFrame
        '''
        if self.path:
            try:
                opt_dict = pd.read_csv(f'{self.path}\\max sharpe and weights.csv').set_index(
                    'Unnamed: 0')['weights'].to_dict()
            except:
                opt_dict = self.optimization()['weights']
        # 股票初始手数（不取整）
        init_shares_dict = {}
        for name, weight in opt_dict.items():
            init_shares_dict[name] = (
                self.funds*weight)/(self.last_price[name]*100)
        ceils = [math.ceil(i) for i in init_shares_dict.values()]
        floors = [math.floor(i) for i in init_shares_dict.values()]
        result = np.array(list(zip(ceils, floors)))
        df_shares = pd.DataFrame(
            list(product(*result)), columns=init_shares_dict.keys())
        return df_shares

    def init_scatter_data(self):
        '''
        初始临近整数组合散点
        '''
        # 初始临近整数组合权重
        df_init_shares = self.init_port()
        for name in self.names:
            df_init_shares[name] = df_init_shares[name]*self.last_price[name]
        df_init_shares['sum'] = df_init_shares.sum(axis=1)
        for name in self.names:
            df_init_shares[name] = df_init_shares[name]/df_init_shares['sum']
        df_init_shares = df_init_shares[df_init_shares['sum']
                                        <= self.funds/100]
        weights = np.array([list(i.values()) for i in list(
            df_init_shares[self.names].T.to_dict().values())])
        # 收益率、风险和夏普比率
        data_dict = self.calculate()
        data_mean = data_dict['mean']
        data_cov = data_dict['cov']
        # 散点DataFrame
        df_init_scatter = pd.DataFrame()
        # 随机权重
        df_init_scatter['weights'] = pd.Series(map(lambda x: str(x), weights))
        # 风险
        df_init_scatter['risk'] = np.sqrt(np.diagonal(
            weights.dot(data_cov).dot(weights.T)))
        # 收益率
        df_init_scatter['rate'] = data_mean.dot(weights.T).T['pctChg']
        # 夏普比率
        df_init_scatter['sharpe'] = (
            df_init_scatter.rate-self.rfr)/df_init_scatter.risk
        df_init_scatter = df_init_scatter.sort_values(
            by='sharpe', ascending=False)
        if self.path:
            makedir(self.path, 'scatter data')
            df_init_scatter.to_csv(
                f'{self.path}\\scatter data\\df_init_scatter.csv')
        return df_init_scatter

    def init_cml(self):
        '''
        初始临近组合散点图
        '''
        if self.path:
            try:
                df_init_scatter = pd.read_csv(
                    f'{self.path}\\scatter data\\df_init_scatter.csv')
                df_boundary_scatter = pd.read_csv(
                    f'{self.path}\\scatter data\\boundary_scatter_data.csv')
            except:
                df_init_scatter = self.init_scatter_data()
                df_boundary_scatter = self.boundary_scatter_data()
        else:
            df_init_scatter = self.init_scatter_data()
            df_boundary_scatter = self.boundary_scatter_data()

        max_sharpe = self.optimization()['sharpe']
        sprint(f'max sharpe: {max_sharpe}')
        plt.style.use('seaborn-paper')
        plt.scatter(df_init_scatter.risk, df_init_scatter.rate,
                    s=100, marker=".", c='r')
        plt.scatter(df_boundary_scatter.risk, df_boundary_scatter.rate,
                    s=10, marker=".", c='b')
        plt.axline(xy1=(0, self.rfr), slope=max_sharpe, c='m')
        plt.xlim(df_init_scatter.risk.min()*0.8,
                 df_init_scatter.risk.max()*1.2)
        plt.ylim(df_init_scatter.rate.min()*0.8,
                 df_init_scatter.rate.max()*1.2)
        plt.xlabel('Risk')
        plt.ylabel('Yield')
        plt.show()

    def init_tree(self):
        '''
        第一次分枝定界结果 --> DataFrame
        '''
        init_tree_list = []
        for index, shares_series in self.init_port().iterrows():
            shares_dict = shares_series.to_dict()
            examed_sharpe = self.exam(shares_dict)
            # if examed_sharpe > 0:
            #     init_tree_list.append([shares_dict, examed_sharpe])
            init_tree_list.append([shares_dict, examed_sharpe])
        # if len(init_tree_list) == 0:
        #     sprint('fund数值过小或组合选择存在问题，存在股票权重为0！！！')
            # raise ValueError('fund数值过小或组合选择存在问题，存在股票权重为0！！！')
        return pd.DataFrame(init_tree_list, columns=['shares', 'sharpe']).sort_values(by='sharpe', ascending=False)

    def near_port_constructor(self, shares_dict={'迎驾贡酒': 1, '明微电子': 2, '健民集团': 3}, near=1):
        '''
        构建临近整数组合 --> DataFrame
        '''
        lists = [shares_dict]
        for name, number in shares_dict.items():
            shares_dict_copy_1 = shares_dict.copy()
            shares_dict_copy_2 = shares_dict.copy()
            if number == 0:
                shares_dict_copy_1[name] = number+near
                lists.append(shares_dict_copy_1)
                continue
            elif number == 1:
                near = 1
            lists.append(shares_dict_copy_1)
            lists.append(shares_dict_copy_2)
            shares_dict_copy_1[name] = number-near
            shares_dict_copy_2[name] = number+near
        return pd.DataFrame(lists)

    def port(self, df_shares, near=1):
        '''
        临近整数组合 --> DataFrame
        '''
        port_list = []
        # 寻找每一个检验剩下的组合的临近组合
        for i in df_shares.itertuples():
            port_list.append(self.near_port_constructor(i.shares, near=near))
        # 拼接所有临近整数解
        return pd.Series(pd.concat(port_list).drop_duplicates().reset_index(drop=True).T.to_dict())

    def tree(self):
        '''
        分枝定界
        返回最优整数解和sharpe
        '''
        # 初始整数组合
        exam_tree = pd.DataFrame()
        exam_tree['weights'] = self.port(self.init_tree(), near=1)
        max_sharpe = -9999999
        sprint('Searching for the integer shares')
        n = 0
        flag = False
        near = 1
        while True:
            n += 1
            tree_list = []
            print(f'第{n}次迭代：')
            for i in tqdm(list(exam_tree.itertuples())):
                examed_sharpe = self.exam(i.weights)
                if examed_sharpe != 0:
                    tree_list.append([i.weights, examed_sharpe])
            df_exam = pd.DataFrame(tree_list, columns=['shares', 'sharpe']).sort_values(
                by='sharpe', ascending=False)
            # 引入过滤条件减少计算量
            df_exam = df_exam[df_exam['sharpe'] >= max_sharpe]
            if len(df_exam) == 1:
                return df_exam.iloc[0].to_dict()
            # 本次迭代最大sharpe
            max_sharpe = df_exam['sharpe'].iloc[0]
            print(
                f'max_sharpe:{max_sharpe}\nshares:{df_exam["shares"].iloc[0]}\n'+'-'*100)
            # 寻找下一个临近点
            if flag:
                near = 1
                flag = False
            elif n > 1:
                near = 2
                flag = True
            exam_tree = pd.DataFrame()
            exam_tree['weights'] = self.port(df_exam, near=near)

    def buy(self):
        '''
        最优整数解
        return a dict contained all above:
        sharpe --> 夏普比率
        weights --> 各股票权重
        cost --> 各股票购买成本
        shares --> 各股票购买手数
        cost --> 总成本
        '''
        result_dict = self.tree()
        result_dict['weights'] = {}
        cost_dict = {}
        for name in self.names:
            cost_dict[name] = result_dict['shares'][name] * \
                self.last_price[name]*100
        sum_cost = np.array(list(cost_dict.values())).sum()
        result_dict['weights'] = {name: cost /
                                  sum_cost for name, cost in cost_dict.items()}
        result_dict['cost'] = cost_dict
        result_dict['sum_cost'] = sum_cost
        if self.path:
            df_opt = pd.DataFrame(result_dict)
            df_opt['sharpe'] = None
            df_opt['sharpe'].iloc[0] = result_dict['sharpe']
            df_opt['sum_cost'] = None
            df_opt['sum_cost'].iloc[0] = result_dict['sum_cost']
            df_opt.columns = ['shares', 'sharpe',
                              'weights', 'cost', 'sum_cost']
            df_opt.to_csv(f'{self.path}\\max sharpe and weights (integer).csv')
        return result_dict

    def weight_tests(self, number=5):
        '''
        构建所有股票个数为number的组合
        '''
        lists = []
        for port in tqdm(list(combinations(self.names, number))):
            self.names = list(port)
            self.data = self.datas.loc[self.names].reset_index(drop=True)
            self.lens = len(self.names)
            test_dict = self.optimization()
            weight_array = np.array(list(test_dict['weights'].values()))
            test_dict['std'] = np.std(weight_array+1)
            test_dict['min'] = np.min(weight_array)
            test_dict['max'] = np.max(weight_array)
            lists.append(test_dict)
            if test_dict['min'] > 0.02:
                sprint(test_dict)
        df_test = pd.DataFrame(
            lists, columns=['weights', 'sharpe', 'std', 'min', 'max'])
        if self.path:
            df_test.to_csv(f'{self.path}\\weight_test.csv', index=False)
        return df_test


class CAPM(Markovitz):

    def __init__(self, names=['比亚迪', '阳光电源', '璞泰来', '紫光国微', '盛新锂能'],
                 start_date='2021-05-01',
                 end_date='2021-11-01',
                 frequency='w',
                 rfr=0.023467,
                 market_index='沪深300指数',
                 path='.\\CAPM cache\\'):
        self.names = names
        self.lens = len(names)
        self.start_date = start_date
        self.end_date = end_date
        self.frequency = frequency
        self.rfr = (rfr*100) / \
            {'d': 365, 'w': 52, 'm': 30}[frequency]
        self.market_index = market_index
        self.path = path
        if self.path:
            makedir(self.path, '')
        sprint('Initializing...')
        if not self.path:
            sd = StockData(names=self.names+[market_index], start_date=self.start_date,
                           end_date=self.end_date, frequency=self.frequency)
            self.datas = sd.stocks_data()
        else:
            try:
                self.datas = pd.read_csv(
                    f'{self.path}\\stock data\\stocks_data.csv')
            except:
                sd = StockData(names=self.names+[market_index], start_date=self.start_date,
                               end_date=self.end_date, frequency=self.frequency, path=self.path)
                self.datas = sd.stocks_data()

        self.datas.index = self.datas['name']
        self.data = self.datas.loc[self.names].reset_index(drop=True)
        self.Rm_data = self.datas.loc[self.market_index].reset_index(drop=True)

    def beta(self, Ri):
        '''
        按照定义计算Beta系数 --> 返回beta值
        '''
        Rm = self.Rm_data['pctChg']
        if len(Ri) != len(Rm):
            raise ValueError('传入的数据长度不等于市场数据长度!')
        return (np.cov(Ri, Rm))[0][1]/np.var(Rm)

    def ls_beta(self, Ri):
        '''
        回归估计beta系数 --> 返回包含alpha和beta的字典
        '''
        y = Ri.reset_index(drop=True)
        x = sm.add_constant(self.Rm_data['pctChg'])
        model = regression.linear_model.OLS(y, x).fit()
        return {'alpha_ols': model.params['const'], 'beta_ols': model.params['pctChg']}

    def scl(self, name='', show=True):
        '''
        给定资产的证券特征线
        '''
        if name not in self.names:
            sprint(
                f'name参数值未给定，或参数值{name}不在给定风险资产范围内！已重新随机选择一种给定风险资产！', color='red')
            name = random.choice(self.names)
        Ri = self.data[self.data.name == name]['pctChg']
        Rm = self.Rm_data['pctChg']
        ls_dict = self.ls_beta(Ri)
        plt.cla()
        plt.axline(xy1=(0, ls_dict['alpha_ols']),
                   slope=ls_dict['beta_ols'], c='m')
        plt.scatter(Ri, Rm, s=10, marker=".", c='b')
        plt.xlabel(f'{name}收益率（%）')
        plt.ylabel(f'{self.market_index}收益率（%）')
        if show:
            plt.show()
        else:
            makedir(self.path, 'scl')
            plt.savefig(f'{self.path}\\scl\\{name}.svg', format='svg')

    def all_beta(self):
        '''
        回归估计beta系数、回归估计截距项alpha的值和按照定义计算beta系数汇总
        '''
        lists = []
        Rm_mean = self.Rm_data['pctChg'].mean()
        for name in self.names:
            pctChg_series = self.data[self.data.name == name]['pctChg']
            dic = self.ls_beta(pctChg_series)
            dic['beta'] = self.beta(pctChg_series)
            dic['Ri'] = pctChg_series.mean()
            dic['Rm'] = Rm_mean
            dic['name'] = name
            lists.append(dic)
        df_all_beta = pd.DataFrame(lists).set_index('name')
        if self.path:
            df_all_beta.to_csv(f'{self.path}\\beta.csv')
        return df_all_beta

    def sml(self, show=True):
        '''
        估计收益率在证券市场线上的分布
        '''
        plt.style.use('seaborn-paper')
        Rm_mean = self.Rm_data['pctChg'].mean()
        k = Rm_mean-self.rfr
        all_beta = self.all_beta()
        plt.cla()
        plt.axline(xy1=(0, self.rfr), slope=k, c='m')
        plt.scatter(all_beta.beta, all_beta.beta*k +
                    self.rfr, s=100, marker=".", c='b')
        plt.scatter(all_beta.beta, all_beta.Ri, s=100, marker=".", c='r')
        for i in all_beta.itertuples():
            plt.annotate(text=i.Index, xy=(
                i.beta, i.Ri), xytext=(i.beta, i.Ri))
            plt.annotate(text=i.Index, xy=(i.beta, i.beta*k +
                         self.rfr), xytext=(i.beta, i.beta*k + self.rfr))
        if all_beta.beta.min() < 0:
            plt.axvline(0, linestyle='--')
        else:
            plt.xlim(0, all_beta.beta.max()*1.2)
        plt.xlabel('Beta')
        plt.ylabel('E(Ri)')
        if show:
            plt.show()
        else:
            plt.savefig(f'{self.path}\\sml.svg', format='svg')

    def all(self):
        '''
        证券市场线 + 所有风险资产的特征线
        '''
        for name in self.names:
            self.scl(name, show=False)
        self.sml(show=False)


if __name__ == '__main__':
    from pprint import pprint
    # 初始化
    capm = CAPM(names=['安科瑞', '紫光国微', '航发控制'],  # 股票组合
                start_date='2020-11-01',  # 开始日期
                end_date='2021-11-01',  # 结束日期
                frequency='w', # 'd' or 'w' or 'm' → 日、周、月 
                rfr=0.023467,  # 无风险利率
                )
        
    # 某种风险资产的收益率时间序列
    Ri=capm.data[capm.data.name == '紫光国微']['pctChg']
        
    # 按照定义计算beta系数 --> 返回beta值            
    pprint(capm.beta(Ri=Ri))

    # 回归估计beta系数 --> 返回包含alpha和beta的字典
    pprint(capm.ls_beta(Ri=Ri))

    # 所有风险资产分别按照【回归估计beta系数、回归估计截距项alpha的值和按照定义计算beta系数】 --> 返回一个DataFrame
    pprint(capm.all_beta())

    # 证券特征线，show=False --> 保存为svg格式矢量图
    capm.scl(name='紫光国微',show=True)

    # 证券市场线，show=False --> 保存为svg格式矢量图
    capm.sml(show=True)

    # 上述所有 --> 保存为svg格式矢量图
    capm.all()



import json
import os
import sys
import time
from functools import reduce
from operator import add
from typing import List, Sequence, Union


import numpy as np
import pandas as pd
from pyecharts import options as opts
from pyecharts.commons.utils import JsCode
from pyecharts.charts import Kline, Line, Bar, Grid

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..data.Stock import StockData
    from ..log.log import hide, show, slog, sprint, makedir
except:
    from data.Stock import StockData
    from log.log import hide, show, slog, sprint, makedir


class SyntheticIndex(object):

    def __init__(self, weights={'贵州茅台': 0.5,
                                '隆基股份': 0.2,
                                '五粮液': 0.3},
                 path='.\\Suluoya cache\\'
                 ):
        self.names = list(weights.keys())
        self.weights = list(weights.values())
        self.lens = len(self.names)
        self.path = path
        makedir(self.path, '')

    def stock_weights(self):
        return dict(zip(self.names, [i/sum(self.weights) for i in self.weights]))

    def stock_data(self):
        if not self.path:
            sd = StockData(names=self.names, path=self.path)
            sd.start_date = sorted(
                [i['ipoDate'] for i in sd.stocks_info().values()])[-1]
            sd.end_date = time.strftime("%Y-%m-%d", time.localtime())
            return sd.stocks_data()[['name', 'date', 'open', 'close', 'low', 'high', 'volume']]
        else:
            try:
                stocks_info = pd.read_csv(
                    f'{self.path}\\stock data\\stocks_info.csv').set_index('name').T.to_dict()
                sd = StockData(names=self.names, path=self.path)
                sd.start_date = sorted([i['ipoDate']
                                       for i in stocks_info.values()])[-1]
                sd.end_date = time.strftime("%Y-%m-%d", time.localtime())
                return sd.stocks_data()[['name', 'date', 'open', 'close', 'low', 'high', 'volume']]
            except:
                sd = StockData(names=self.names, path=self.path)
                sd.start_date = sorted(
                    [i['ipoDate'] for i in sd.stocks_info().values()])[-1]
                sd.end_date = time.strftime("%Y-%m-%d", time.localtime())
                return sd.stocks_data()[['name', 'date', 'open', 'close', 'low', 'high', 'volume']]

    def get_data(self):
        data = self.stock_data()
        data['weight'] = data['name'].map(self.stock_weights())
        data['real_weight'] = data.weight/data.open
        date = data.date.unique()
        x = [list(data[data.date == i].real_weight /
                  data[data.date == i].real_weight.sum()) for i in date]
        data = data.sort_values(by=['date', 'name'])
        data['z_weight'] = reduce(add, tuple(x))
        data.date = data.date.astype(str)
        data.open = (data.open*data.z_weight).map(lambda x: round(x, 2))
        data.close = (data.close*data.z_weight).map(lambda x: round(x, 2))
        data.low = (data.low*data.z_weight).map(lambda x: round(x, 2))
        data.high = (data.high*data.z_weight).map(lambda x: round(x, 2))
        data.volume = (data.volume*data.z_weight).map(lambda x: round(x, 0))
        data = data.groupby('date').sum()
        data['date'] = data.index
        return data[['date', 'open', 'close', 'low', 'high', 'volume']]

    def calculate(self):
        data = self.get_data()
        # 计算EMA(12)和EMA(16)
        data['EMA12'] = data['close'].ewm(alpha=2 / 13, adjust=False).mean()
        data['EMA26'] = data['close'].ewm(alpha=2 / 27, adjust=False).mean()
        # 计算DIFF、DEA、MACD
        data['DIFF'] = data['EMA12'] - data['EMA26']
        data['DEA'] = data['DIFF'].ewm(alpha=2 / 10, adjust=False).mean()
        data['MACD'] = 2 * (data['DIFF'] - data['DEA'])
        data['DIFF'] = data['DIFF'].map(lambda x: round(x, 2))
        data['DEA'] = data['DEA'].map(lambda x: round(x, 2))
        data['MACD'] = data['MACD'].map(lambda x: round(x, 2))
        # 上市首日，DIFF、DEA、MACD均为0
        data['DIFF'].iloc[0] = 0
        data['DEA'].iloc[0] = 0
        data['MACD'].iloc[0] = 0
        data['datas'] = data[['open', 'close', 'low', 'high', 'volume',
                              'EMA12', 'EMA26', 'DIFF', 'DEA', 'MACD']].values.tolist()
        return data

    def calculate_ma(self, day_count, data: int):
        result: List[Union[float, str]] = []
        for i in range(len(data["date"])):
            if i < day_count:
                result.append("-")
                continue
            sum_total = 0.0
            for j in range(day_count):
                sum_total += float(data["datas"][i - j][1])
            result.append(abs(float("%.2f" % (sum_total / day_count))))
        return result

    def kline(self):
        data = self.calculate().to_dict('list')
        kline = (
            Kline()
            .add_xaxis(xaxis_data=data["date"])
            .add_yaxis(
                series_name="portfolio index",
                y_axis=data["datas"],
                itemstyle_opts=opts.ItemStyleOpts(
                    color="#ef232a",
                    color0="#14b143",
                    border_color="#ef232a",
                    border_color0="#14b143",
                ),
            )
            .set_global_opts(
                xaxis_opts=opts.AxisOpts(
                    type_="category",
                    is_scale=True,
                    boundary_gap=False,
                    axisline_opts=opts.AxisLineOpts(is_on_zero=False),
                    splitline_opts=opts.SplitLineOpts(is_show=False),
                    split_number=20,
                    min_="dataMin",
                    max_="dataMax",
                ),
                yaxis_opts=opts.AxisOpts(
                    is_scale=True, splitline_opts=opts.SplitLineOpts(is_show=True)
                ),
                datazoom_opts=[
                    opts.DataZoomOpts(
                        is_show=False, type_="inside", xaxis_index=[0, 0], range_end=100
                    ),
                    opts.DataZoomOpts(
                        is_show=True, xaxis_index=[0, 1], pos_top="97%", range_end=100
                    ),
                    opts.DataZoomOpts(is_show=False, xaxis_index=[
                                      0, 2], range_end=100),
                ],
                tooltip_opts=opts.TooltipOpts(
                    trigger="axis",
                    axis_pointer_type="cross",
                    background_color="rgba(245, 245, 245, 0.8)",
                    border_width=1,
                    border_color="#ccc",
                    textstyle_opts=opts.TextStyleOpts(color="#000"),
                ),
                brush_opts=opts.BrushOpts(
                    x_axis_index="all",
                    brush_link="all",
                    out_of_brush={"colorAlpha": 0.1},
                    brush_type="lineX",
                ),
                # 三个图的 axis 连在一块
                axispointer_opts=opts.AxisPointerOpts(
                    is_show=True,
                    link=[{"xAxisIndex": "all"}],
                    label=opts.LabelOpts(background_color="#777"),
                ),
            )
        )

        kline_line = (
            Line()
            .add_xaxis(xaxis_data=data["date"])
            .add_yaxis(
                series_name="MA5",
                y_axis=self.calculate_ma(day_count=5, data=data),
                is_smooth=True,
                linestyle_opts=opts.LineStyleOpts(opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
            )
            .add_yaxis(
                series_name="MA10",
                y_axis=self.calculate_ma(day_count=10, data=data),
                is_smooth=True,
                linestyle_opts=opts.LineStyleOpts(opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
            )
            .add_yaxis(
                series_name="MA20",
                y_axis=self.calculate_ma(day_count=20, data=data),
                is_smooth=True,
                linestyle_opts=opts.LineStyleOpts(opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
            )
            .add_yaxis(
                series_name="MA30",
                y_axis=self.calculate_ma(day_count=30, data=data),
                is_smooth=True,
                linestyle_opts=opts.LineStyleOpts(opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
            )
            .set_global_opts(
                xaxis_opts=opts.AxisOpts(
                    type_="category",
                    grid_index=1,
                    axislabel_opts=opts.LabelOpts(is_show=False),
                ),
                yaxis_opts=opts.AxisOpts(
                    grid_index=1,
                    split_number=3,
                    axisline_opts=opts.AxisLineOpts(is_on_zero=False),
                    axistick_opts=opts.AxisTickOpts(is_show=False),
                    splitline_opts=opts.SplitLineOpts(is_show=False),
                    axislabel_opts=opts.LabelOpts(is_show=True),
                ),
            )
        )
        # Overlap Kline + Line
        overlap_kline_line = kline.overlap(kline_line)

        # Bar-1
        bar_1 = (
            Bar()
            .add_xaxis(xaxis_data=data["date"])
            .add_yaxis(
                series_name="Volumn",
                y_axis=data["volume"],
                xaxis_index=1,
                yaxis_index=1,
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(
                    color=JsCode(
                        """
                    function(params) {
                        var colorList;
                        if (barData[params.dataIndex][1] > barData[params.dataIndex][0]) {
                            colorList = '#ef232a';
                        } else {
                            colorList = '#14b143';
                        }
                        return colorList;
                    }
                    """
                    )
                ),
            )
            .set_global_opts(
                xaxis_opts=opts.AxisOpts(
                    type_="category",
                    grid_index=1,
                    axislabel_opts=opts.LabelOpts(is_show=False),
                ),
                legend_opts=opts.LegendOpts(is_show=False),
            )
        )

        # Bar-2 (Overlap Bar + Line)
        bar_2 = (
            Bar()
            .add_xaxis(xaxis_data=data["date"])
            .add_yaxis(
                series_name="MACD",
                y_axis=data["MACD"],
                xaxis_index=2,
                yaxis_index=2,
                label_opts=opts.LabelOpts(is_show=False),
                itemstyle_opts=opts.ItemStyleOpts(
                    color=JsCode(
                        """
                            function(params) {
                                var colorList;
                                if (params.data >= 0) {
                                colorList = '#ef232a';
                                } else {
                                colorList = '#14b143';
                                }
                                return colorList;
                            }
                            """
                    )
                ),
            )
            .set_global_opts(
                xaxis_opts=opts.AxisOpts(
                    type_="category",
                    grid_index=2,
                    axislabel_opts=opts.LabelOpts(is_show=False),
                ),
                yaxis_opts=opts.AxisOpts(
                    grid_index=2,
                    split_number=4,
                    axisline_opts=opts.AxisLineOpts(is_on_zero=False),
                    axistick_opts=opts.AxisTickOpts(is_show=False),
                    splitline_opts=opts.SplitLineOpts(is_show=False),
                    axislabel_opts=opts.LabelOpts(is_show=True),
                ),
                legend_opts=opts.LegendOpts(is_show=False),
            )
        )

        line_2 = (
            Line()
            .add_xaxis(xaxis_data=data["date"])
            .add_yaxis(
                series_name="DIFF",
                y_axis=data["DIFF"],
                xaxis_index=2,
                yaxis_index=2,
                label_opts=opts.LabelOpts(is_show=False),
            )
            .add_yaxis(
                series_name="DEA",
                y_axis=data["DEA"],
                xaxis_index=2,
                yaxis_index=2,
                label_opts=opts.LabelOpts(is_show=False),
            )
            .set_global_opts(legend_opts=opts.LegendOpts(is_show=False))
        )
        # 最下面的柱状图和折线图
        overlap_bar_line = bar_2.overlap(line_2)

        # 最后的 Grid
        grid_chart = Grid(init_opts=opts.InitOpts(
            width="1500px", height="750px"))

        # 这个是为了把 data.datas 这个数据写入到 html 中,还没想到怎么跨 series 传值
        # demo 中的代码也是用全局变量传的
        grid_chart.add_js_funcs("var barData = {}".format(data["datas"]))

        # K线图和 MA5 的折线图
        grid_chart.add(
            overlap_kline_line,
            grid_opts=opts.GridOpts(
                pos_left="3%", pos_right="1%", height="60%"),
        )
        # Volumn 柱状图
        grid_chart.add(
            bar_1,
            grid_opts=opts.GridOpts(
                pos_left="3%", pos_right="1%", pos_top="71%", height="10%"
            ),
        )
        # MACD DIFS DEAS
        grid_chart.add(
            overlap_bar_line,
            grid_opts=opts.GridOpts(
                pos_left="3%", pos_right="1%", pos_top="82%", height="14%"
            ),
        )
        makedir(self.path, 'kline')
        grid_chart.render(
            f"{self.path}\\kline\\{str(self.names)}_{time.strftime('%Y-%m-%d', time.localtime())}.html")


if __name__ == "__main__":
    si = SyntheticIndex()
    si.draw_chart()
